
#pragma warning(disable:4996)

#include "shoutAkiRa.h"
#include "ShoutPlayer.h"
#include "RadioTuner.h"

#include <algorithm>

#include "buffer.h"
#include "html.h"
#include "mkdirs.h"
#include "encoding/jstring.h"
#include "CombinePath.h"

#include <shlobj.h>


enum Band {
	UNKNOWN_BAND,
	AM = 1,
	FM = 2,
};


struct tune_t {
	double frequency;
	std::string description;
	tune_t()
		: frequency(0)
	{
	}
	tune_t(double frequency, std::string const &desc)
		: frequency(frequency)
		, description(desc)
	{
	}
	bool operator < (tune_t const &r) const
	{
		return frequency < r.frequency;
	}
};

struct ShoutShark::Privates {
	httpserver::HTTP_Server httpserver;
	ShoutPlayer shoutplayer;
	RadioTuner radiotuner;
	std::string my_program_files_dir;

	std::vector<std::string> com_ports;

	int radio_device_index;
	std::string control_com_port;

	std::string scserver_address;
	int scserver_port;
	std::string scserver_password;

	struct {
		Band band;
		double frequency;
	} current_tune;

	std::vector<tune_t> am_presets;
	std::vector<tune_t> fm_presets;

	Privates()
	{
		scserver_port = 0;
		radio_device_index = 0;
		control_com_port = "\\\\.\\COM1";
	}
};

ShoutShark::ShoutShark()
{
	privates = new Privates();

	char tmp[MAX_PATH];
	GetModuleFileName(0, tmp, MAX_PATH);
	char *p = strrchr(tmp, '\\');
	if (p) {
		*p = 0;
		p = strrchr(tmp, '\\');
		if (p) {
			*p = 0;
		}
	}
	privates->my_program_files_dir = tmp;
}

ShoutShark::~ShoutShark()
{
	delete privates;
}

void ShoutShark::TuneAM(double kHz)
{
	privates->current_tune.band = AM;
	privates->current_tune.frequency = kHz;
	privates->radiotuner.tune_am((int)kHz);
}

void ShoutShark::TuneFM(double MHz)
{
	privates->current_tune.band = FM;
	privates->current_tune.frequency = MHz;
	privates->radiotuner.tune_fm(MHz);
}

typedef std::pair<std::string, std::string> string_pair_t;
typedef std::vector<string_pair_t> tiny_dictionary_t;

bool getvalue(tiny_dictionary_t const &dic, std::string const &key, std::string *val)
{
	for (std::vector<string_pair_t>::const_iterator it = dic.begin(); it != dic.end(); it++) {
		if (it->first == key) {
			if (val) {
				*val = it->second;
			}
			return true;
		}
	}
	return false;
}

inline bool hasvalue(tiny_dictionary_t const &dic, std::string const &key)
{
	return getvalue(dic, key, 0);
}

inline std::string getvalue(tiny_dictionary_t const &dic, std::string const &key, std::string const &def)
{
	std::string val;
	if (getvalue(dic, key, &val)) {
		return val;
	}
	return def;
}

inline std::string getvalue(tiny_dictionary_t const &dic, std::string const &key)
{
	std::string val;
	getvalue(dic, key, &val);
	return val;
}

void decode_uri_parameter(std::string *v)
{
	if (strchr(v->c_str(), '+') || strchr(v->c_str(), '%')) {
		std::vector<char> buffer;
		buffer.reserve(v->size());
		size_t i = 0;
		while (i < v->size()) {
			char c = v->at(i);
			i++;
			if (c == '+') {
				buffer.push_back(' ');
			} else if (c == '%') {
				if (i + 1 < v->size()) {
					char tmp[3] = { v->at(i), v->at(i + 1), 0 };
					buffer.push_back((char)strtol(tmp, 0, 16));
				}
				i += 2;
			} else {
				buffer.push_back(c);
			}
		}
		v->assign(&buffer[0], buffer.size());
	}
}

void parse_http_query(char const *str, tiny_dictionary_t *dic)
{
	dic->clear();
	char const *ptr = str;
	char const *end = str + strlen(str);
	char const *left = ptr;
	char const *eq = 0;
	while (1) {
		if (*ptr == '=') {
			if (!eq) {
				eq = ptr;
			}
		} else if (*ptr == '&' || ptr == end) {
			char const *right = ptr;
			std::string k;
			std::string v;
			if (eq) {
				k.assign(left, eq);
				v.assign(eq + 1, right);
			} else {
				k.assign(left, right);
			}
			decode_uri_parameter(&v);
			dic->push_back(std::pair<std::string, std::string>(k, v));
			if (ptr == end) {
				break;
			}
			left = ptr + 1;
			eq = 0;
		}
		ptr++;
	}
}

std::string utf8_wstring_to_string(std::wstring const &str)
{
	return soramimi::encoding::convert(soramimi::encoding::UTF8, str);
}

void ShoutShark::SetRadioDevice(int device, std::string const &comport)
{
	privates->shoutplayer.Close();

	if (device >= 0) {
		privates->shoutplayer.Open(device);
		privates->shoutplayer.Start();
	}

	privates->radiotuner.close();
	privates->radiotuner.open(comport.c_str());

	privates->radio_device_index = device;
	privates->control_com_port = comport;

	std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/device");
	FILE *fp = fopen(path.c_str(), "w");
	if (fp) {
		fprintf(fp, "%d\t%s\n"
			, privates->radio_device_index
			, privates->control_com_port.c_str()
			);
		fclose(fp);
	}
}

void ShoutShark::SetServerInfo(std::string const &address, int port, std::string const &password, bool restart)
{
	privates->shoutplayer.SetServerInfo(address.c_str(), port, password.c_str());

	privates->scserver_address = address;
	privates->scserver_port = port;
	privates->scserver_password = password;

	if (restart) {
		privates->shoutplayer.Open(privates->radio_device_index);
		privates->shoutplayer.Start();
	}

	std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/scserver");
	FILE *fp = fopen(path.c_str(), "w");
	if (fp) {
		fprintf(fp, "%s\t%u\t%s"
			, privates->scserver_address.c_str()
			, privates->scserver_port
			, privates->scserver_password.c_str()
			);
		fclose(fp);
	}
}

void ShoutShark::SavePresets()
{
	std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/presets");
	FILE *fp = fopen(path.c_str(), "w");
	if (fp) {
		for (std::vector<tune_t>::const_iterator it = privates->am_presets.begin(); it != privates->am_presets.end(); it++) {
			fprintf(fp, "AM\t%u\t%s\n"
				, (int)it->frequency
				, it->description.c_str()
				);
		}
		for (std::vector<tune_t>::const_iterator it = privates->fm_presets.begin(); it != privates->fm_presets.end(); it++) {
			fprintf(fp, "FM\t%.2f\t%s\n"
				, (double)it->frequency
				, it->description.c_str()
				);
		}
		fclose(fp);
	}
}

std::string ShoutShark::GetDescription()
{
	if (privates->current_tune.band == AM) {
		for (std::vector<tune_t>::const_iterator it = privates->am_presets.begin(); it != privates->am_presets.end(); it++) {
			if (privates->current_tune.frequency == it->frequency) {
				return it->description;
			}
		}
	} else if (privates->current_tune.band == FM) {
		for (std::vector<tune_t>::const_iterator it = privates->fm_presets.begin(); it != privates->fm_presets.end(); it++) {
			if (privates->current_tune.frequency == it->frequency) {
				return it->description;
			}
		}
	}
	return std::string();
}

static inline std::string tostring(int n)
{
	char tmp[100];
	sprintf(tmp, "%u", n);
	return tmp;
}

http_status_t const *ShoutShark::http_get_handler(std::string const &uri, httpserver::http_request_t *request, httpserver::http_response_t *response)
{
	tiny_dictionary_t query;

	{
		char const *header = "/shoutakira";
		int n = strlen(header);
		if (strncmp(uri.c_str(), header, n) == 0) {
			if (uri[n] == 0) {
				goto accept;
			}
			if (uri[n] == '?') {
				char const *q = uri.c_str() + 12;
				parse_http_query(q, &query);
				goto accept;
			}
		}
	}

	return 0;

accept:

	if (getvalue(query, "mode") == "tune") {
		std::string band = getvalue(query, "band");
		double f = strtod(getvalue(query, "f").c_str(), 0);

		char text[100];
		text[0] = 0;

		if (stricmp(band.c_str(), "am") == 0) {
			TuneAM(f);
			sprintf(text, "AM%u", (int)f);
		} else if (stricmp(band.c_str(), "fm") == 0) {
			TuneFM(f);
			sprintf(text, "FM%.2f", (double)f);
		} else {
			return response->set_location("/shoutakira");
		}

		std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/tune");
		FILE *fp = fopen(path.c_str(), "w");
		if (fp) {
			fputs(text, fp);
			fclose(fp);
		}

		return response->set_location("/shoutakira");
	}


	response->set_content_type("text/html");

	string_buffer<char> buffer;

	buffer.write("<html>\n");
	buffer.write("<head>\n");
	buffer.write("<title>shoutAkiRa</title>\n");
	buffer.write("<meta http-equiv='Content-Type' content='text/html; charset=UTF-8'>\n");
	buffer.write("<link rel='stylesheet' type='text/css' href='/shoutakira.css'>\n");

	buffer.write("</head>\n");
	buffer.write("<body>\n");

	buffer.write("<h1 id=header><a href='/shoutakira'>shoutAkiRa</a></h1>\n");

	if (!hasvalue(query, "mode")) {
		buffer.write("<h2 id=status>Status</h2>\n");
		buffer.write("<div>\n");
		{
			char tmp[100];
			tmp[0] = 0;
			if (privates->current_tune.band == AM) {
				sprintf(tmp, "<div>%ukHz AM</div>\n", (int)privates->current_tune.frequency);
			} else if (privates->current_tune.band == FM) {
				sprintf(tmp, "<div>%.2fMHz FM</div>\n", (double)privates->current_tune.frequency);
			}
			buffer.write(tmp);
		}
		{
			std::string desc = GetDescription();
			buffer.write("<div>\n");
			buffer.write(encode_html_entity(desc));
			buffer.write("</div>\n");
		}
		buffer.write("</div>\n");
		buffer.write("<h2 id=presets>Presets</h2>\n");
		buffer.write("<table>\n");
		buffer.write("<tr>\n");
		buffer.write("<th colspan=2>AM</th>\n");
		buffer.write("</tr>\n");

		for (std::vector<tune_t>::const_iterator it = privates->am_presets.begin(); it != privates->am_presets.end(); it++) {
			char url[100];
			sprintf(url, "/shoutakira?mode=tune&band=am&f=%u", (int)it->frequency);
			char tmp[1000];
			sprintf(tmp, 
				"<tr>\n"
				"<td align=right><a href='%s'>%ukHz</a></td>\n"
				"<td><a href='%s'>"
				, url
				, (int)it->frequency
				, url
				);
			buffer.write(tmp);
			buffer.write(encode_html_entity(it->description));
			sprintf(tmp,
				"<a href='%s'></td>\n</tr>\n\n"
				, url
				);
			buffer.write(tmp);
		}

		buffer.write("<tr>\n");
		buffer.write("<th colspan=2>FM</th>\n");
		buffer.write("</tr>\n");

		for (std::vector<tune_t>::const_iterator it = privates->fm_presets.begin(); it != privates->fm_presets.end(); it++) {
			char url[100];
			sprintf(url, "/shoutakira?mode=tune&band=fm&f=%.2f", (double)it->frequency);
			char tmp[1000];
			sprintf(tmp, 
				"<tr>\n"
				"<td align=right><a href='%s'>%0.2fMHz</a></td>\n"
				"<td><a href='%s'>"
				, url
				, (double)it->frequency
				, url
				);
			buffer.write(tmp);
			buffer.write(encode_html_entity(it->description));
			sprintf(tmp,
				"<a href='%s'></td>\n</tr>\n\n"
				, url
				);
			buffer.write(tmp);
		}

		buffer.write("</table>\n");

		buffer.write("<a href='/shoutakira?mode=insertpreset'>Insert into presets</a><br>\n");
		buffer.write("<a href='/shoutakira?mode=deletepreset'>Delete from presets</a><br>\n");

		buffer.write("<h2 id=manualtune>Manual Tune</h2>\n");

		buffer.write("<form method=get action='/shoutakira'>\n");
		buffer.write("<input type=hidden name=mode value=tune>\n");
		buffer.write("Band ");
		buffer.write("<input type=radio name=band value=am>AM (kHz)</input>");
		buffer.write("<input type=radio name=band value=fm>FM (MHz)</input>");
		buffer.write("<br>\n");
		buffer.write("Frequency <input type=text name=f><br>\n");
		buffer.write("<input type=submit value='Tune'><br>\n");
		buffer.write("</form>\n");

		buffer.write("<h2 id=radiodevice>Radio Device</h2>\n");
		buffer.write("<p><a href='/shoutakira?mode=configdevice'>Configure radio device.</a></p>\n");

		buffer.write("<h2 id=scserver>SHOUTcast Server</h2>\n");
		{
			std::string url = std::string("http://") + privates->scserver_address + ":" + tostring(privates->scserver_port) + "/";
			buffer.write(std::string("<p><a href='") + url + "'>" + encode_html_entity(url) + "</p>\n");
		}
		buffer.write("<p><a href='/shoutakira?mode=configserver'>Configure SHOUTcast server.</a></p>\n");

	} else if (getvalue(query, "mode") == "insertpreset") {

		if (hasvalue(query, "submit")) {
			std::string tune = getvalue(query, "tune");
			std::string desc = getvalue(query, "desc");
			std::vector<tune_t> *p = 0;
			double freq;
			if (sscanf(tune.c_str(), "AM%lf", &freq) == 1) {
				p = &privates->am_presets;
			} else if (sscanf(tune.c_str(), "FM%lf", &freq) == 1) {
				p = &privates->fm_presets;
			}
			if (p) {
				size_t i = p->size();
				while (i > 0) {
					i--;
					if (freq == p->at(i).frequency) {
						p->erase(p->begin() + i);
					}
				}
				p->push_back(tune_t(freq, desc));
				std::sort(p->begin(), p->end());
			}
			SavePresets();
			return response->set_location("/shoutakira");
		}

		buffer.write("<h2 id=insertpreset>Insert into presets</h2>\n");

		{
			char tmp[100];
			if (privates->current_tune.band == AM) {
				sprintf(tmp, "%ukHz AM ", (int)privates->current_tune.frequency);
			} else if (privates->current_tune.band == FM) {
				sprintf(tmp, "%0.2fMHz FM ", (double)privates->current_tune.frequency);
			} else {
				return response->set_location("/shoutakira");
			}
			buffer.write("<p>");
			buffer.write(tmp);
			buffer.write("</p>\n");
		}

		buffer.write("<h4 id=description>Description</h4>\n");

		buffer.write("<form method=get action='/shoutakira'>\n");
		buffer.write("<input type=hidden name=mode value=insertpreset>\n");
		{
			char tmp[100];
			tmp[0] = 0;
			if (privates->current_tune.band == AM) {
				sprintf(tmp, "<input type=hidden name=tune value='AM%u'>\n", (int)privates->current_tune.frequency);
			} else if (privates->current_tune.band ==FM) {
				sprintf(tmp, "<input type=hidden name=tune value='FM%.2f'>\n", (double)privates->current_tune.frequency);
			}
			buffer.write(tmp);
		}
		buffer.write(std::string("<input type=text name=desc value='") + encode_html_entity(GetDescription()) + "'>\n");
		buffer.write("<input type=submit name=submit value='Insert'><br>\n");
		buffer.write("</form>\n");

	} else if (getvalue(query, "mode") == "deletepreset") {

		if (hasvalue(query, "submit")) {
			std::string tune = getvalue(query, "tune");
			std::vector<tune_t> *p = 0;
			double freq;
			if (sscanf(tune.c_str(), "AM%lf", &freq) == 1) {
				p = &privates->am_presets;
			} else if (sscanf(tune.c_str(), "FM%lf", &freq) == 1) {
				p = &privates->fm_presets;
			}
			if (p) {
				size_t i = p->size();
				while (i > 0) {
					i--;
					if (freq == p->at(i).frequency) {
						p->erase(p->begin() + i);
					}
				}
			}
			SavePresets();
			return response->set_location("/shoutakira");
		}

		buffer.write("<h2 id=deletepreset>Delete from presets</h2>\n");

		{
			char tmp[100];
			if (privates->current_tune.band == AM) {
				sprintf(tmp, "%ukHz AM : ", (int)privates->current_tune.frequency);
			} else if (privates->current_tune.band == FM) {
				sprintf(tmp, "%0.2fMHz FM : ", (double)privates->current_tune.frequency);
			} else {
				return response->set_location("/shoutakira");
			}
			buffer.write("<p>");
			buffer.write(tmp);
			buffer.write(GetDescription());
			buffer.write("</p>\n");
		}
		buffer.write("<form method=get action='/shoutakira'>");
		buffer.write("<input type=hidden name=mode value=deletepreset>\n");
		{
			char tmp[100];
			tmp[0] = 0;
			if (privates->current_tune.band == AM) {
				sprintf(tmp, "<input type=hidden name=tune value='AM%u'>\n", (int)privates->current_tune.frequency);
			} else if (privates->current_tune.band ==FM) {
				sprintf(tmp, "<input type=hidden name=tune value='FM%.2f'>\n", (double)privates->current_tune.frequency);
			}
			buffer.write(tmp);
		}
		buffer.write("<input type=submit name=submit value='Delete'><br>\n");
		buffer.write("</form>\n");

	} else if (getvalue(query, "mode") == "configdevice") {

		if (hasvalue(query, "submit")) {
			int device = atoi(getvalue(query, "device").c_str());
			int index = atoi(getvalue(query, "comport").c_str());
			std::string port;
			if (index >= 0 && index < (int)privates->com_ports.size()) {
				port = privates->com_ports[index];
			}
			SetRadioDevice(device, port);
			return response->set_location("/shoutakira");
		}

		buffer.write("<h2 id=radiodevice>Radio Device</h2>\n");
		buffer.write("<p>Select input device and control port.</p>\n");

		buffer.write("<form name=f>\n");
		buffer.write("<input type=hidden name=mode value=configdevice>\n");

		buffer.write("<table>\n");
		buffer.write("<tr>\n");
		buffer.write("<td>Input Device</td><td><select name=device>\n");
		unsigned int n = waveInGetNumDevs();
		for (unsigned int i = 0; i < n; i++) {
			char tmp[200];
			WAVEINCAPSW caps;
			waveInGetDevCapsW(i, &caps, sizeof(caps));
			sprintf(tmp, "<option value=%u>%s</option>\n"
				, i
				, utf8_wstring_to_string(encode_html_entity(caps.szPname)).c_str()
				);
			buffer.write(tmp);
		}
		buffer.write("</select></td>\n");
		buffer.write("</tr>\n");

		buffer.write("<tr>\n");
		buffer.write("<td>Control Port</td><td><select name=comport>\n");
		{
			for (int i = 0; i < (int)privates->com_ports.size(); i++) {
				char tmp[100];
				char const *p = privates->com_ports[i].c_str();
				if (strncmp(p, "\\\\.\\", 4) == 0) {
					p += 4;
				}
				sprintf(tmp, "<option value=%u>%s</option>\n", i, p);
				buffer.write(tmp);
			}
		}
		buffer.write("</select></td>\n");
		buffer.write("</tr>\n");

		buffer.write("<tr>\n");
		buffer.write("<td></td><td><input type=submit name=submit value='Apply'></td>\n");
		buffer.write("</tr>\n");

		buffer.write("</table>\n");

		buffer.write("</form>\n");

	} else if (getvalue(query, "mode") == "configserver") {

		if (hasvalue(query, "submit")) {
			std::string address = getvalue(query, "address");
			std::string port = getvalue(query, "port");
			std::string password = getvalue(query, "password");
			SetServerInfo(address, atoi(port.c_str()), password, true);
			return response->set_location("/shoutakira");
		}

		buffer.write("<h2 id=scserver>Configure SHOUTcast Server</h2>\n");
		buffer.write("<p>Input server address and port.</p>\n");
		buffer.write("<form method=get action='/shoutakira'>\n");
		buffer.write("<input type=hidden name=mode value=configserver>\n");
		buffer.write("<table>\n");
		buffer.write("<tr>\n");
		buffer.write(std::string("<td>Address</td><td><input type=text name=address value='") + encode_html_entity(privates->scserver_address) + "'></td>\n");
		buffer.write("</tr>\n");
		buffer.write("<tr>\n");
		buffer.write(std::string("<td>Port</td><td><input type=text name=port value='") + tostring(privates->scserver_port) + "'></td>\n");
		buffer.write("</tr>\n");
		buffer.write("<tr>\n");
		buffer.write(std::string("<td>Password</td><td><input type=password name=password value='") + encode_html_entity(privates->scserver_password) + "'></td>\n");
		buffer.write("</tr>\n");
		buffer.write("<tr>\n");
		buffer.write("<td></td><td><input type=submit name=submit value='Apply'></td>\n");
		buffer.write("</tr>\n");
		buffer.write("</table>\n");
		buffer.write("</form>\n");
	}

	buffer.write("<hr>\n");
	buffer.write("<small id=footer>\n");
	buffer.write("<div><a href='http://www.soramimi.jp/shoutakira/'>shoutAkiRa</a> v0.0, the streaming data source server for SHOUTcast.</div>\n");
	buffer.write("<div>Copyright (C) 2010 S.Fuchita.</div>\n");
	buffer.write("</small>\n");

	buffer.write("</body>\n");
	buffer.write("</html>\n");

	response->buffer.insert(response->buffer.end(), buffer.get(), buffer.get() + buffer.size());
	return http_200_ok;
}

http_status_t const *ShoutShark::http_get_handler_(void *cookie, std::string const &uri, httpserver::http_request_t *request, httpserver::http_response_t *response)
{
	ShoutShark *me = (ShoutShark *)cookie;
	return me->http_get_handler(uri, request, response);
}

void initialize_winsock()
{
	WSADATA wsaData;
	WORD wVersionRequested = MAKEWORD(1, 1);
	WSAStartup(wVersionRequested, &wsaData);
	atexit((void (*)(void))(WSACleanup));
}

extern SERVICE_STATUS_HANDLE service_status_handle;
extern SERVICE_STATUS service_status;

BOOL EnumSerialPorts(std::vector<std::string> *ports)
{
	//Make sure we clear out any elements which may already be in the array
	ports->clear();

	//Up to 255 COM ports are supported so we iterate through all of them seeing
	//if we can open them or if we fail to open them, get an access denied or general error error.
	//Both of these cases indicate that there is a COM port at that number. 
	for (UINT i=1; i<256; i++) {
		//Form the Raw device name
		std::string sPort;
		{
			char tmp[100];
			sprintf(tmp, "\\\\.\\COM%d", i);
			sPort = tmp;
		}

		//Try to open the port
		BOOL bSuccess = FALSE;
		HANDLE hPort = ::CreateFile(sPort.c_str(), GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);
		if (hPort == INVALID_HANDLE_VALUE) {
			DWORD dwError = GetLastError();

			//Check to see if the error was because some other app had the port open or a general failure
			if (dwError == ERROR_ACCESS_DENIED || dwError == ERROR_GEN_FAILURE || dwError == ERROR_SHARING_VIOLATION || dwError == ERROR_SEM_TIMEOUT)
				bSuccess = TRUE;
		} else {
			//The port was opened successfully
			bSuccess = TRUE;

			//Don't forget to close the port, since we are going to do nothing with it anyway
			CloseHandle(hPort);
		}

		//Add the port number to the array which will be returned
		if (bSuccess) {
			ports->push_back(sPort);
		}
	}

	//Return the success indicator
	return TRUE;
}

void ShoutShark::Main(bool as_service)
{
	initialize_winsock();

	if (as_service) {
		service_status.dwCurrentState = SERVICE_START_PENDING;
		service_status.dwCheckPoint = 0;
		service_status.dwWaitHint = 1000;
		SetServiceStatus(service_status_handle, &service_status);
	}

	EnumSerialPorts(&privates->com_ports);

	{
		std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/device");
		FILE *fp = fopen(path.c_str(), "r");
		if (fp) {
			char tmp[100];
			fgets(tmp, 100, fp);
			int d, t;
			if (sscanf(tmp, "%d", &d, &t) == 1) {
				privates->radio_device_index = d;
			}
			fclose(fp);
		}
	}

	{
		std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/scserver");
		FILE *fp = fopen(path.c_str(), "r");
		if (fp) {
			bool success = false;
			std::string host;
			int port;
			std::string password;
			char tmp[100];
			fgets(tmp, 100, fp);
			char const *ptr = tmp;
			char const *end = ptr + strlen(ptr);
			char const *tab;
			tab = strchr(ptr, '\t');
			if (tab) {
				host.assign(ptr, tab);
				for (ptr = tab; ptr < end && isspace(*ptr); ptr++);
				port = atoi(ptr);
				tab = strchr(ptr, '\t');
				if (tab) {
					for (ptr = tab; ptr < end && isspace(*ptr); ptr++);
					while (ptr < end && isspace(end[-1] & 0xff)) {
						end--;
					}
					password.assign(ptr, end);
					success = true;
				}
			}
			fclose(fp);

			if (success) {
				SetServerInfo(host.c_str(), port, password.c_str(), false);
			}
		}
	}

	{
		std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/presets");
		FILE *fp = fopen(path.c_str(), "r");
		if (fp) {
			char tmp[1000];
			while (fgets(tmp, 1000, fp)) {
				if (strncmp(tmp, "AM\t", 3) == 0) {
					char const *ptr = strchr(tmp + 3, '\t');
					if (ptr) {
						ptr++;
						char const *end = ptr + strlen(ptr);
						while (ptr < end && isspace(end[-1] & 0xff)) {
							end--;
						}
						double f = atoi(tmp + 3);
						privates->am_presets.push_back(tune_t(f, std::string(ptr, end)));
					}
				} else if (strncmp(tmp, "FM\t", 3) == 0) {
					char const *ptr = strchr(tmp + 3, '\t');
					if (ptr) {
						ptr++;
						char const *end = ptr + strlen(ptr);
						while (ptr < end && isspace(end[-1] & 0xff)) {
							end--;
						}
						double f = strtod(tmp + 3, 0);
						privates->fm_presets.push_back(tune_t(f, std::string(ptr, end)));
					}
				}
			}
			std::sort(privates->am_presets.begin(), privates->am_presets.end());
			std::sort(privates->fm_presets.begin(), privates->fm_presets.end());
			fclose(fp);
		}
	}

	int wwwport = 8080;
	{
		std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/wwwport");
		FILE *fp = fopen(path.c_str(), "r");
		if (fp) {
			char tmp[100];
			if (fgets(tmp, 100, fp)) {
				wwwport = atoi(tmp);
			}
			fclose(fp);
		}
	}
	privates->httpserver.set_tcp_port(wwwport);
	privates->httpserver.set_document_root(CombinePath(privates->my_program_files_dir.c_str(), "www"));
	privates->httpserver.add_http_get_handler(http_get_handler_, this);
	privates->httpserver.start();

	privates->shoutplayer.Open(privates->radio_device_index);
	privates->shoutplayer.Start();

	privates->radiotuner.open(privates->control_com_port.c_str());

	{
		Band band = UNKNOWN_BAND;
		double frequency = 0;

		std::string path = CombinePath(privates->my_program_files_dir.c_str(), "settings/tune");
		FILE *fp = fopen(path.c_str(), "r");
		if (fp) {
			char tmp[100];
			if (fgets(tmp, 100, fp)) {
				double f;
				if (sscanf(tmp, "AM%lf", &f) == 1) {
					TuneAM(f);
				} else if (sscanf(tmp, "FM%lf", &f) == 1) {
					TuneFM(f);
				}
			}
			fclose(fp);
		}
	}

	if (as_service) {
		service_status.dwCurrentState = SERVICE_RUNNING;
		service_status.dwCheckPoint = 0;
		service_status.dwWaitHint = 0;
		SetServiceStatus(service_status_handle, &service_status);
	}

	while (1) {
		if (as_service) {
			if (service_status.dwCurrentState == SERVICE_STOP_PENDING) {
				break;
			}
		}
		Sleep(1000);
	}

	privates->radiotuner.close();

	privates->shoutplayer.Stop();
	privates->shoutplayer.Close();

	privates->httpserver.stop();
	privates->httpserver.join();

	if (as_service) {
		service_status.dwCurrentState = SERVICE_STOPPED;
		SetServiceStatus(service_status_handle, &service_status);
	}
}

